@@ -16,7 +16,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: command_reply.c,v 1.106 2002/06/22 20:53:16 priikone Exp $ */
+/* $Id: command_reply.c,v 1.112 2002/11/02 11:45:45 priikone Exp $ */
 /*
  * Command reply functions are "the otherside" of the command functions.
  * Reply to a command sent by server is handled by these functions.
@@ -241,13 +241,22 @@ silc_client_command_reply_whois_save(SilcClientCommandReplyContext cmd,
     client_entry->fingerprint_len = fingerprint_len;
   }
 
+  /* Take Requested Attributes if set. */
+  tmp = silc_argument_get_arg_type(cmd->args, 11, &len);
+  if (tmp) {
+    if (client_entry->attrs)
+      silc_attribute_payload_list_free(client_entry->attrs);
+    client_entry->attrs = silc_attribute_payload_parse(tmp, len);
+  }
+
   client_entry->status &= ~SILC_CLIENT_STATUS_RESOLVING;
 
   /* Notify application */
   if (!cmd->callbacks_count && notify)
-    COMMAND_REPLY((ARGS, client_entry, nickname, username, realname, 
+    COMMAND_REPLY((SILC_ARGS, client_entry, nickname, username, realname, 
 		   has_channels ? &channels : NULL, mode, idle, 
-		   fingerprint, has_user_modes ? &ch_user_modes : NULL));
+		   fingerprint, has_user_modes ? &ch_user_modes : NULL,
+		   client_entry->attrs));
 }
 
 /* Received reply for WHOIS command. This maybe called several times
@@ -339,7 +348,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(whowas)
   /* Notify application. We don't save any history information to any
      cache. Just pass the data to the application for displaying on 
      the screen. */
-  COMMAND_REPLY((ARGS, client_entry, nickname, username, realname));
+  COMMAND_REPLY((SILC_ARGS, client_entry, nickname, username, realname));
 
   /* Pending callbacks are not executed if this was an list entry */
   if (cmd->status != SILC_STATUS_OK &&
@@ -413,7 +422,7 @@ silc_client_command_reply_identify_save(SilcClientCommandReplyContext cmd,
 
     /* Notify application */
     if (notify)
-      COMMAND_REPLY((ARGS, client_entry, name, info));
+      COMMAND_REPLY((SILC_ARGS, client_entry, name, info));
     break;
 
   case SILC_ID_SERVER:
@@ -440,7 +449,7 @@ silc_client_command_reply_identify_save(SilcClientCommandReplyContext cmd,
 
     /* Notify application */
     if (notify)
-      COMMAND_REPLY((ARGS, server_entry, name, info));
+      COMMAND_REPLY((SILC_ARGS, server_entry, name, info));
     break;
 
   case SILC_ID_CHANNEL:
@@ -462,7 +471,7 @@ silc_client_command_reply_identify_save(SilcClientCommandReplyContext cmd,
 
     /* Notify application */
     if (notify)
-      COMMAND_REPLY((ARGS, channel_entry, name, info));
+      COMMAND_REPLY((SILC_ARGS, channel_entry, name, info));
     break;
   }
 
@@ -561,18 +570,19 @@ SILC_CLIENT_CMD_REPLY_FUNC(nick)
   /* Take the new nickname too */
   tmp = silc_argument_get_arg_type(cmd->args, 3, &len);
   if (tmp) {
-    silc_idcache_del_by_context(conn->client_cache, conn->local_entry);
+    silc_idcache_del_by_context(conn->internal->client_cache,
+				conn->local_entry);
     if (conn->nickname)
       silc_free(conn->nickname);
     conn->nickname = strdup(tmp);
     conn->local_entry->nickname = conn->nickname;
     silc_client_nickname_format(cmd->client, conn, conn->local_entry);
-    silc_idcache_add(conn->client_cache, strdup(tmp),
+    silc_idcache_add(conn->internal->client_cache, strdup(tmp),
                      conn->local_entry->id, conn->local_entry, 0, NULL);
   }
     
   /* Notify application */
-  COMMAND_REPLY((ARGS, conn->local_entry, conn->local_entry->nickname));
+  COMMAND_REPLY((SILC_ARGS, conn->local_entry, conn->local_entry->nickname));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_NICK);
@@ -630,7 +640,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(list)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, channel_entry, name, topic, usercount));
+  COMMAND_REPLY((SILC_ARGS, channel_entry, name, topic, usercount));
 
   /* Pending callbacks are not executed if this was an list entry */
   if (cmd->status != SILC_STATUS_OK &&
@@ -694,7 +704,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(topic)
   }
   
   /* Notify application */
-  COMMAND_REPLY((ARGS, channel, topic));
+  COMMAND_REPLY((SILC_ARGS, channel, topic));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_TOPIC);
@@ -740,7 +750,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(invite)
   tmp = silc_argument_get_arg_type(cmd->args, 3, &len);
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, channel, tmp));
+  COMMAND_REPLY((SILC_ARGS, channel, tmp));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_INVITE);
@@ -762,7 +772,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(kill)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_KILL);
@@ -822,7 +832,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(info)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, server, server->server_name, server->server_info));
+  COMMAND_REPLY((SILC_ARGS, server, server->server_name, server->server_info));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_INFO);
@@ -830,6 +840,38 @@ SILC_CLIENT_CMD_REPLY_FUNC(info)
   silc_client_command_reply_free(cmd);
 }
 
+/* Received reply to STATS command. */
+
+SILC_CLIENT_CMD_REPLY_FUNC(stats)
+{
+  SilcClientCommandReplyContext cmd = (SilcClientCommandReplyContext)context;
+  SilcClientConnection conn = (SilcClientConnection)cmd->sock->user_data;
+  unsigned char *tmp, *buf = NULL;
+  SilcUInt32 len, buf_len = 0;
+
+  if (cmd->error != SILC_STATUS_OK) {
+    SAY(cmd->client, conn, SILC_CLIENT_MESSAGE_ERROR,
+	"%s", silc_get_status_message(cmd->error));
+    COMMAND_REPLY_ERROR;
+    goto out;
+  }
+
+  /* Get server ID */
+  tmp = silc_argument_get_arg_type(cmd->args, 2, &len);
+  if (!tmp)
+    goto out;
+
+  /* Get statistics structure */
+  buf = silc_argument_get_arg_type(cmd->args, 3, &buf_len);
+
+  /* Notify application */
+  COMMAND_REPLY((SILC_ARGS, buf, buf_len));
+
+ out:
+  SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_STATS);
+  silc_client_command_reply_free(cmd);
+}
+
 /* Received reply to PING command. The reply time is shown to user. */
 
 SILC_CLIENT_CMD_REPLY_FUNC(ping)
@@ -850,26 +892,26 @@ SILC_CLIENT_CMD_REPLY_FUNC(ping)
   curtime = time(NULL);
   id = silc_id_str2id(cmd->packet->src_id, cmd->packet->src_id_len,
 		      cmd->packet->src_id_type);
-  if (!id || !conn->ping) {
+  if (!id || !conn->internal->ping) {
     COMMAND_REPLY_ERROR;
     goto out;
   }
 
-  for (i = 0; i < conn->ping_count; i++) {
-    if (!conn->ping[i].dest_id)
+  for (i = 0; i < conn->internal->ping_count; i++) {
+    if (!conn->internal->ping[i].dest_id)
       continue;
-    if (SILC_ID_SERVER_COMPARE(conn->ping[i].dest_id, id)) {
-      diff = curtime - conn->ping[i].start_time;
+    if (SILC_ID_SERVER_COMPARE(conn->internal->ping[i].dest_id, id)) {
+      diff = curtime - conn->internal->ping[i].start_time;
       SAY(cmd->client, conn, SILC_CLIENT_MESSAGE_INFO, 
 	  "Ping reply from %s: %d second%s", 
-	  conn->ping[i].dest_name, diff, 
+	  conn->internal->ping[i].dest_name, diff, 
 	  diff == 1 ? "" : "s");
       
-      conn->ping[i].start_time = 0;
-      silc_free(conn->ping[i].dest_id);
-      conn->ping[i].dest_id = NULL;
-      silc_free(conn->ping[i].dest_name);
-      conn->ping[i].dest_name = NULL;
+      conn->internal->ping[i].start_time = 0;
+      silc_free(conn->internal->ping[i].dest_id);
+      conn->internal->ping[i].dest_id = NULL;
+      silc_free(conn->internal->ping[i].dest_name);
+      conn->internal->ping[i].dest_name = NULL;
       break;
     }
   }
@@ -877,7 +919,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(ping)
   silc_free(id);
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_PING);
@@ -1054,7 +1096,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(join)
     silc_client_save_channel_key(cmd->client, conn, keyp, channel);
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, channel_name, channel, mode, 0, 
+  COMMAND_REPLY((SILC_ARGS, channel_name, channel, mode, 0, 
 		 keyp ? keyp->head : NULL, NULL,
 		 NULL, topic, hmac, list_count, client_id_list, 
 		 client_mode_list));
@@ -1105,7 +1147,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(motd)
     while(cp[i] != 0) {
       if (cp[i++] == '\n') {
 	memset(line, 0, sizeof(line));
-	strncat(line, cp, i - 1);
+	silc_strncat(line, sizeof(line), cp, i - 1);
 	cp += i;
 	
 	if (i == 2)
@@ -1121,7 +1163,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(motd)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, motd));
+  COMMAND_REPLY((SILC_ARGS, motd));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_MOTD);
@@ -1154,7 +1196,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(umode)
   conn->local_entry->mode = mode;
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, mode));
+  COMMAND_REPLY((SILC_ARGS, mode));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_UMODE);
@@ -1209,7 +1251,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(cmode)
   channel->mode = mode;
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, channel, mode));
+  COMMAND_REPLY((SILC_ARGS, channel, mode));
 
   silc_free(channel_id);
 
@@ -1292,7 +1334,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(cumode)
     chu->mode = mode;
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, mode, channel, client_entry));
+  COMMAND_REPLY((SILC_ARGS, mode, channel, client_entry));
   silc_free(client_id);
   silc_free(channel_id);
   
@@ -1314,7 +1356,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(kick)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_KICK);
@@ -1334,7 +1376,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(silcoper)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_SILCOPER);
@@ -1354,7 +1396,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(oper)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_OPER);
@@ -1375,7 +1417,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(detach)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
   /* Generate the detachment data and deliver it to the client in the
      detach client operation */
@@ -1404,7 +1446,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(watch)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_WATCH);
@@ -1448,7 +1490,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(ban)
   tmp = silc_argument_get_arg_type(cmd->args, 3, &len);
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, channel, tmp));
+  COMMAND_REPLY((SILC_ARGS, channel, tmp));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_BAN);
@@ -1492,7 +1534,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(leave)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS, channel));
+  COMMAND_REPLY((SILC_ARGS, channel));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_LEAVE);
@@ -1681,7 +1723,7 @@ silc_client_command_reply_users_save(SilcClientCommandReplyContext cmd,
 
   /* Notify application */
   if (notify)
-    COMMAND_REPLY((ARGS, channel, list_count, &client_id_list, 
+    COMMAND_REPLY((SILC_ARGS, channel, list_count, &client_id_list, 
 		   &client_mode_list));
 
  out:
@@ -1779,7 +1821,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(getkey)
     }
 
     /* Notify application */
-    COMMAND_REPLY((ARGS, id_type, client_entry, public_key));
+    COMMAND_REPLY((SILC_ARGS, id_type, client_entry, public_key));
   } else if (id_type == SILC_ID_SERVER) {
     /* Received server's public key */
     server_id = silc_id_payload_get_id(idp);
@@ -1790,7 +1832,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(getkey)
     }
 
     /* Notify application */
-    COMMAND_REPLY((ARGS, id_type, server_entry, public_key));
+    COMMAND_REPLY((SILC_ARGS, id_type, server_entry, public_key));
   }
 
  out:
@@ -2023,7 +2065,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(connect)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_PRIV_CONNECT);
@@ -2043,7 +2085,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(close)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_PRIV_CLOSE);
@@ -2063,7 +2105,7 @@ SILC_CLIENT_CMD_REPLY_FUNC(shutdown)
   }
 
   /* Notify application */
-  COMMAND_REPLY((ARGS));
+  COMMAND_REPLY((SILC_ARGS));
 
  out:
   SILC_CLIENT_PENDING_EXEC(cmd, SILC_COMMAND_PRIV_SHUTDOWN);
