@@ -17,7 +17,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: idlist.c,v 1.68 2002/05/17 21:16:18 priikone Exp $ */
+/* $Id: idlist.c,v 1.78 2002/11/04 14:03:59 priikone Exp $ */
 
 #include "serverincludes.h"
 #include "idlist.h"
@@ -76,19 +76,27 @@ void silc_idlist_del_data(void *entry)
     silc_hash_free(idata->hash);
   if (idata->public_key)
     silc_pkcs_public_key_free(idata->public_key);
+
+  idata->send_key = NULL;
+  idata->receive_key = NULL;
+  idata->rekey = NULL;
+  idata->hmac_send = NULL;
+  idata->hmac_receive = NULL;
+  idata->hash = NULL;
+  idata->public_key = NULL;
 }
 
 /* Purges ID cache */
 
 SILC_TASK_CALLBACK_GLOBAL(silc_idlist_purge)
 {
+  SilcServer server = app_context;
   SilcIDListPurge i = (SilcIDListPurge)context;
 
-  SILC_LOG_DEBUG(("Start"));
+  SILC_LOG_DEBUG(("Purging cache"));
 
   silc_idcache_purge(i->cache);
-  silc_schedule_task_add(i->schedule, 0, 
-			 silc_idlist_purge,
+  silc_schedule_task_add(server->schedule, 0, silc_idlist_purge,
 			 (void *)i, i->timeout, 0,
 			 SILC_TASK_TIMEOUT, SILC_TASK_PRI_LOW);
 }
@@ -151,13 +159,13 @@ silc_idlist_find_server_by_id(SilcIDList id_list, SilcServerID *id,
 
   server = (SilcServerEntry)id_cache->context;
 
-  if (ret_entry)
-    *ret_entry = id_cache;
-
   if (server && registered && 
       !(server->data.status & SILC_IDLIST_STATUS_REGISTERED))
     return NULL;
 
+  if (ret_entry)
+    *ret_entry = id_cache;
+
   SILC_LOG_DEBUG(("Found"));
 
   return server;
@@ -179,13 +187,13 @@ silc_idlist_find_server_by_name(SilcIDList id_list, char *name,
 
   server = (SilcServerEntry)id_cache->context;
   
-  if (ret_entry)
-    *ret_entry = id_cache;
-
   if (server && registered &&
       !(server->data.status & SILC_IDLIST_STATUS_REGISTERED))
     return NULL;
 
+  if (ret_entry)
+    *ret_entry = id_cache;
+
   SILC_LOG_DEBUG(("Found"));
 
   return server;
@@ -231,13 +239,13 @@ silc_idlist_find_server_by_conn(SilcIDList id_list, char *hostname,
   
   silc_idcache_list_free(list);
 
-  if (ret_entry)
-    *ret_entry = id_cache;
-
   if (server && registered &&
       !(server->data.status & SILC_IDLIST_STATUS_REGISTERED))
     return NULL;
 
+  if (ret_entry)
+    *ret_entry = id_cache;
+
   SILC_LOG_DEBUG(("Found"));
 
   return server;
@@ -282,13 +290,14 @@ silc_idlist_replace_server_id(SilcIDList id_list, SilcServerID *old_id,
 
 int silc_idlist_del_server(SilcIDList id_list, SilcServerEntry entry)
 {
-  SILC_LOG_DEBUG(("Start"));
-
   if (entry) {
     /* Remove from cache */
     if (!silc_idcache_del_by_context(id_list->servers, entry))
       return FALSE;
 
+    SILC_LOG_DEBUG(("Deleting server %s", entry->server_name ?
+		    entry->server_name : ""));
+
     /* Free data */
     silc_free(entry->server_name);
     silc_free(entry->id);
@@ -359,12 +368,15 @@ int silc_idlist_del_client(SilcIDList id_list, SilcClientEntry entry)
     if (!silc_idcache_del_by_context(id_list->clients, entry))
       return FALSE;
 
+    assert(!silc_hash_table_count(entry->channels));
+
     /* Free data */
     silc_free(entry->nickname);
     silc_free(entry->servername);
     silc_free(entry->username);
     silc_free(entry->userinfo);
     silc_free(entry->id);
+    silc_free(entry->attrs);
     silc_hash_table_free(entry->channels);
 
     memset(entry, 'F', sizeof(*entry));
@@ -482,13 +494,13 @@ silc_idlist_find_client_by_id(SilcIDList id_list, SilcClientID *id,
 
   client = (SilcClientEntry)id_cache->context;
 
-  if (ret_entry)
-    *ret_entry = id_cache;
-
   if (client && registered &&
       !(client->data.status & SILC_IDLIST_STATUS_REGISTERED))
     return NULL;
 
+  if (ret_entry)
+    *ret_entry = id_cache;
+
   SILC_LOG_DEBUG(("Found"));
 
   return client;
@@ -559,13 +571,15 @@ void silc_idlist_client_destructor(SilcIDCache cache,
 
   client = (SilcClientEntry)entry->context;
   if (client) {
+    assert(!silc_hash_table_count(client->channels));
     silc_free(client->nickname);
     silc_free(client->username);
     silc_free(client->userinfo);
     silc_free(client->id);
+    silc_free(client->attrs);
     silc_hash_table_free(client->channels);
 
-    memset(client, 'F', sizeof(*client));
+    memset(client, 'A', sizeof(*client));
     silc_free(client);
   }
 }
@@ -587,7 +601,7 @@ silc_idlist_add_channel(SilcIDList id_list, char *channel_name, int mode,
 {
   SilcChannelEntry channel;
 
-  SILC_LOG_DEBUG(("Adding new channel entry"));
+  SILC_LOG_DEBUG(("Adding new channel %s", channel_name));
 
   channel = silc_calloc(1, sizeof(*channel));
   channel->channel_name = channel_name;
@@ -596,7 +610,7 @@ silc_idlist_add_channel(SilcIDList id_list, char *channel_name, int mode,
   channel->router = router;
   channel->channel_key = channel_key;
   channel->hmac = hmac;
-  channel->created = time(0);
+  channel->created = channel->updated = time(0);
   if (!channel->hmac)
     if (!silc_hmac_alloc(SILC_DEFAULT_HMAC, NULL, &channel->hmac)) {
       silc_free(channel);
@@ -626,7 +640,8 @@ static void silc_idlist_del_channel_foreach(void *key, void *context,
   SilcChannelClientEntry chl = (SilcChannelClientEntry)context;
 
   SILC_LOG_DEBUG(("Removing client %s from channel %s",
-		  chl->client->nickname, chl->channel->channel_name));
+		  chl->client->nickname ? chl->client->nickname :
+		  (unsigned char *)"", chl->channel->channel_name));
 
   /* Remove the context from the client's channel hash table as that
      table and channel's user_list hash table share this same context. */
@@ -638,13 +653,13 @@ static void silc_idlist_del_channel_foreach(void *key, void *context,
 
 int silc_idlist_del_channel(SilcIDList id_list, SilcChannelEntry entry)
 {
-  SILC_LOG_DEBUG(("Start"));
-
   if (entry) {
     /* Remove from cache */
     if (!silc_idcache_del_by_context(id_list->channels, entry))
       return FALSE;
 
+    SILC_LOG_DEBUG(("Deleting channel %s", entry->channel_name));
+
     /* Free all client entrys from the users list. The silc_hash_table_free
        will free all the entries so they are not freed at the foreach 
        callback. */
@@ -656,6 +671,8 @@ int silc_idlist_del_channel(SilcIDList id_list, SilcChannelEntry entry)
     silc_free(entry->channel_name);
     silc_free(entry->id);
     silc_free(entry->topic);
+    silc_free(entry->invite_list);
+    silc_free(entry->ban_list);
     if (entry->channel_key)
       silc_cipher_free(entry->channel_key);
     if (entry->key) {
@@ -697,6 +714,9 @@ silc_idlist_find_channel_by_name(SilcIDList id_list, char *name,
 
   SILC_LOG_DEBUG(("Found"));
 
+  /* Touch channel */
+  ((SilcChannelEntry)id_cache->context)->updated = time(NULL);
+
   return id_cache->context;
 }
 
@@ -725,6 +745,9 @@ silc_idlist_find_channel_by_id(SilcIDList id_list, SilcChannelID *id,
 
   SILC_LOG_DEBUG(("Found"));
 
+  /* Touch channel */
+  channel->updated = time(NULL);
+
   return channel;
 }
 
@@ -761,6 +784,9 @@ silc_idlist_replace_channel_id(SilcIDList id_list, SilcChannelID *old_id,
 
   SILC_LOG_DEBUG(("Replaced"));
 
+  /* Touch channel */
+  channel->updated = time(NULL);
+
   return channel;
 }
 
